@@ -8,6 +8,8 @@
 import os
 from .utils import *
 
+import optumi_core as optumi
+
 
 class ProgressMessage:
     def __init__(self, updates: list = []):
@@ -28,10 +30,8 @@ class Summary:
         self._running_lines = running_lines
 
     def download(self, path: str = None):
-        f_name = os.path.abspath(
-            os.path.expanduser(
-                self._name.split("/")[-1] + ".summary" if path is None else path
-            )
+        f_name = optumi.utils.normalize_path(
+            self._name.split("/")[-1] + ".summary" if path is None else path
         )
         with open(f_name, "w+") as f:
             f.write(
