@@ -12,9 +12,11 @@ from .CloudFileVersion import CloudFileVersion
 from .Log import Log
 from .Summary import Summary
 from .Program import Program
+from .Machine import Machine
 from .utils import *
 
 import json, time, os, re
+from sys import exit
 
 from enum import Enum
 
@@ -33,20 +35,6 @@ class Progress(Enum):
     DETAIL = "detail"
 
 
-class Workloads:
-    @classmethod
-    def list(cls, status: Status = None):
-        user_information = json.loads(optumi.core.get_user_information(True).text)
-
-        workloads = []
-
-        # Add apps from user information if they don't already exist
-        if "jobs" in user_information:
-            for app_map in user_information["jobs"]:
-                workloads.append(Workload(*Workload.reconstruct(app_map)))
-        return workloads
-
-
 class Workload:
     def __init__(
         self,
@@ -64,6 +52,8 @@ class Workload:
         output: list = [],
         num_patches: int = 0,
         output_files: list = [],
+        machine: Machine = None,
+        token: str = None,
     ):
         self._path = path
         self._program = program
@@ -71,6 +61,10 @@ class Workload:
         self._workload_uuid = workload_uuid
         self._module_uuid = module_uuid
 
+        self._nb_config = nb_config
+
+        self._run_num = run_num
+
         self._initializing_lines = initializing_lines.copy()
         self._preparing_lines = preparing_lines.copy()
         self._running_lines = running_lines.copy()
@@ -82,7 +76,8 @@ class Workload:
 
         self._output_files = output_files.copy()
 
-        self._run_num = run_num
+        self._machine = machine
+        self._token = token
 
         self._last_refresh = time.time()
 
@@ -123,6 +118,8 @@ class Workload:
         output_files = []
         num_patches = 0
         program = None
+        machine = None
+        token = None
 
         # Add modules
         for module in app_map["modules"]:
@@ -160,16 +157,18 @@ class Workload:
             if "notebook" in module:
                 program = module["notebook"]
 
-            # var mod: Module = new Module(module.uuid, module.machine ? Machine.parse(module.machine) : null, module.token, output, updates, outputFiles, monitoring, lastPatch);
-            # mod.applyPatch = app.applyPatch;
-            # app._modules.append(mod);
+            if "machine" in module:
+                machine = Machine(*Machine.reconstruct(module["machine"]))
+
+            if "token" in module:
+                token = module["token"]
 
         return (
             app_map["name"],
             program,
             app_map["uuid"],
             module_uuid,
-            app_map["nbConfig"],
+            json.loads(app_map["nbConfig"]),
             app_map["runNum"],
             initializing,
             preparing,
@@ -179,6 +178,8 @@ class Workload:
             output,
             num_patches,
             output_files,
+            machine,
+            token,
         )
 
     def _refresh(self):
@@ -205,6 +206,8 @@ class Workload:
                             self._output,
                             self._last_patches,
                             self._output_files,
+                            self._machine,
+                            self._token,
                         ) = Workload.reconstruct(app_map)
 
     # def __print_status(self):
@@ -234,7 +237,9 @@ class Workload:
             ):
                 if progress == Progress.SUMMARY:
                     print(
-                        update[0] if update[0].endswith("\n") else update[0] + "\n",
+                        update[0]
+                        if update[0].endswith("\n")
+                        else self.__adjust_message(update[0]) + "\n",
                         end="",
                         flush=True,
                     )
@@ -280,49 +285,63 @@ class Workload:
 
             self.__print_updates(progress, updates_to_print)
 
-            updates = json.loads(
-                optumi.pull_module_status_updates(
-                    [self._workload_uuid],
-                    [self._module_uuid],
-                    [len(self._update_lines)],
-                    [len(self._output)],
-                    [0],
-                    [self._last_patches],
-                ).content
-            )[self._module_uuid]
+            if self._module_uuid:
+                updates = json.loads(
+                    optumi.pull_module_status_updates(
+                        [self._workload_uuid],
+                        [self._module_uuid],
+                        [len(self._update_lines)],
+                        [len(self._output)],
+                        [0],
+                        [self._last_patches],
+                    ).content
+                )[self._module_uuid]
 
-            new_updates = updates["updates"]
-            new_updates_mod = updates["updatesmod"]
-            for i in range(len(new_updates)):
-                self._update_lines.append((new_updates[i], new_updates_mod[i]))
+                new_updates = updates["updates"]
+                new_updates_mod = updates["updatesmod"]
+                for i in range(len(new_updates)):
+                    self._update_lines.append((new_updates[i], new_updates_mod[i]))
 
-            output_to_print = []
+                output_to_print = []
 
-            new_output = updates["output"]
-            new_output_mod = updates["outputmod"]
-            for i in range(len(new_output)):
-                update = new_output[i]
-                self._output.append((update, new_output_mod[i]))
-                output_to_print.append(self._output[-1])
+                new_output = updates["output"]
+                new_output_mod = updates["outputmod"]
+                for i in range(len(new_output)):
+                    update = new_output[i]
+                    self._output.append((update, new_output_mod[i]))
+                    output_to_print.append(self._output[-1])
 
-            self.__print_output(progress, output_to_print)
+                self.__print_output(progress, output_to_print)
 
-            if "patches" in updates:
-                new_patches = updates["patches"]
-                self._last_patches += len(new_patches)
+                if "patches" in updates:
+                    new_patches = updates["patches"]
+                    self._last_patches += len(new_patches)
 
-            self._output_files = CloudStorage(
-                [
-                    CloudFileVersion(
-                        updates["files"][i],
-                        updates["hashes"][i],
-                        int(updates["filessize"][i]),
-                        updates["filescrt"][i],
-                        updates["filesmod"][i],
+                self._output_files = CloudStorage(
+                    [
+                        CloudFileVersion(
+                            updates["files"][i],
+                            updates["hashes"][i],
+                            int(updates["filessize"][i]),
+                            updates["filescrt"][i],
+                            updates["filesmod"][i],
+                        )
+                        for i in range(len(updates["files"]))
+                    ]
+                )
+
+                if "machine" in updates:
+                    self._machine = Machine(*Machine.reconstruct(updates["machine"]))
+
+                if self._token == None and "token" in updates:
+                    self._token = updates["token"]
+                    print(
+                        "Session available at: https://"
+                        + self._machine.dns_name
+                        + ":54321/?token="
+                        + self._token
                     )
-                    for i in range(len(updates["files"]))
-                ]
-            )
+                    return
 
             if self.__get_status() == Status.COMPLETED:
                 break
@@ -333,15 +352,37 @@ class Workload:
         #     print()
 
     def stop(self, wait=True):
-        optumi.core.stop_notebook(self._workload_uuid, self._module_uuid)
-        if wait and self.status == Status.RUNNING:
-            self.wait(progress=Progress.SILENT)
+        from .Workloads import Workloads
+
+        if (
+            optumi.login.is_dynamic()
+            and Workloads.current()._workload_uuid == self._workload_uuid
+        ):
+            print("Stopping current workload")
+            # This is for sessions
+            code = os.system("jupyter lab stop 54321")
+            if code > 15:  # 15 is SIGTERM
+                # This is for jobs
+                exit(0)
+        else:
+            print("Stopping workload " + self.name + "...")
+            if self.status != Status.COMPLETED:
+                optumi.core.stop_notebook(
+                    self._workload_uuid, None
+                )  # Remove the second argument when we move to 3.14
+                if wait and self.status == Status.RUNNING:
+                    self.wait(progress=Progress.SILENT)
+            print("...completed")
 
     def remove(self, wait=True):
-        optumi.core.stop_notebook(self._workload_uuid, self._module_uuid)
-        if wait and self.status == Status.RUNNING:
-            self.wait(progress=Progress.SILENT)
-        optumi.core.teardown_notebook(self._workload_uuid, self._module_uuid)
+        # TODO: Add an override flag to stop workload before removing
+        if self.status != Status.COMPLETED:
+            self.stop()
+        print("Removing workload " + self.name + "...")
+        optumi.core.teardown_notebook(
+            self._workload_uuid, None
+        )  # Remove the second argument when we move to 3.14
+        print("...completed")
 
     @property
     def status(self):
@@ -377,6 +418,26 @@ class Workload:
 
         return Status.LAUNCHING
 
+    def __adjust_message(self, message):
+        # We will say a session is starting until we can connect to it
+        if (
+            self._nb_config["interactive"]
+            and message == "Running"
+            and self._token == None
+        ):
+            return "Connecting"
+        # We call a running app 'Connected'
+        if self._nb_config["interactive"] and message == "Running":
+            return "Connected"
+        # We call a terminated or completed app 'closed',
+        if self._nb_config["interactive"] and message == "Terminating":
+            return "Closing"
+        if self._nb_config["interactive"] and message == "Terminated":
+            return "Closed"
+        if self._nb_config["interactive"] and message == "Completed":
+            return "Closed"
+        return message
+
     @property
     def detailed_status(self):
         self._refresh()
@@ -484,7 +545,11 @@ class Workload:
 
     @property
     def path(self):
-        return os.path.abspath(os.path.expanduser(self._path))
+        return optumi.utils.normalize_path(self._path, strict=False)
+
+    @property
+    def machine(self):
+        return self._machine
 
     def __str__(self):
         return str(self.name)
