@@ -14,7 +14,9 @@ from .CloudStorage import *
 from .Server import *
 from .Resource import *
 from .Workload import *
+from .Workloads import *
 from .Machine import *
+from .Machines import *
 from .Notifications import *
 from .Notebook import *
 from .Container import *
